package eu.jonahbauer.raytracing.scene;

import eu.jonahbauer.raytracing.math.Octree;
import eu.jonahbauer.raytracing.math.Range;
import eu.jonahbauer.raytracing.math.Ray;
import eu.jonahbauer.raytracing.math.Vec3;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;

public final class Scene implements Hittable {
    private final @NotNull Octree<@NotNull Hittable> octree;
    private final @NotNull List<@NotNull Hittable> list;

    public Scene(@NotNull List<? extends @NotNull Hittable> objects) {
        this.octree = newOctree(objects);
        this.list = new ArrayList<>();

        for (Hittable object : objects) {
            var bbox = object.getBoundingBox();
            if (bbox.isPresent()) {
                octree.add(bbox.get(), object);
            } else {
                list.add(object);
            }
        }
    }

    @Override
    public @NotNull Optional<HitResult> hit(@NotNull Ray ray, @NotNull Range range) {
        var state = new State();
        state.range = range;

        octree.hit(ray, object -> hit(state, ray, object));
        list.forEach(object -> hit(state, ray, object));

        return Optional.ofNullable(state.result);
    }

    private boolean hit(@NotNull State state, @NotNull Ray ray, @NotNull Hittable object) {
        var r = object.hit(ray, state.range);
        if (r.isPresent()) {
            if (state.range.surrounds(r.get().t())){
                state.result = r.get();
                state.range = new Range(state.range.min(), state.result.t());
            }
            return true;
        } else {
            return false;
        }
    }

    private static @NotNull Octree<Hittable> newOctree(@NotNull List<? extends Hittable> objects) {
        Vec3 center = Vec3.ZERO, max = Vec3.MIN, min = Vec3.MAX;

        int i = 1;
        for (Hittable object : objects) {
            var bbox = object.getBoundingBox();
            if (bbox.isPresent()) {
                center = Vec3.average(center, bbox.get().center(), i++);
                max = Vec3.max(max, bbox.get().max());
                min = Vec3.min(min, bbox.get().min());
            }
        }

        var dimension = Arrays.stream(new double[] {
                Math.abs(max.x() - center.x()),
                Math.abs(max.y() - center.y()),
                Math.abs(max.z() - center.z()),
                Math.abs(min.x() - center.x()),
                Math.abs(min.y() - center.y()),
                Math.abs(min.z() - center.z())
        }).max().orElse(10);

        return new Octree<>(center, dimension);
    }

    private static class State {
        HitResult result;
        Range range;
    }
}
